module array;


private import std.string;


/**
  List of all features:

  we will assume v = Array!(T) for these examples.

  v.count     - number of items in this Array
  v.capacity  - amount this Array can hold without reallocating
  v.empty     - true if count==0

  v.min       - minimum valid index
  v.max       - maximum valid index

  v.clear()   - clears Array
  v.reverse() - reverses Array

  v[]         - returns dynamic T array of all elements(reference, not copy)

  v.append(T)   - pushes a value on this Array, returns index
  v.append(T[]) - pushes an array of values
  v.insert(T)   - inserts a value at front of array, same as insertAt(0,T)
  v.insertAt(N,T) - inserts a value T at a specified position N

  v.pointer(N)  - pointer to member at N

  v << T ;    - same as push(int)
  v << T[]    - same as push(int[])

  v.pop()     - removes last item added, returns it
  v >> a      - same as pop(), stores value in a

  v.first()   - returns first item added
  v.last()    - returns last item added

  v.first(N)  - returns first N items added
  v.last(N)   - returns last N items added

  v.next()    - equivalent to pushing a new item, returns a pointer to it

  v.remove(T) - removes the element which == T
  v.removaAll(T) - removes all elements which == T
  v.removeAt(N) - removes a specified element, later elements indices shift down by 1
  v.removeRange(N,N2) - removes N..N2, after completion arr[N2] is in the N position

  v.drop(n)   - removes the last n items added to Array, == removeAt(-1)
  v.shift(n)  - removes the first N items added to the Array, == removeRange(0,N)

  v.swap(N1,N2) - exchanges the values at N1 and N2

  v.find(a)   - finds the first occurrence of a in Array or -1
  v.find(a,start) - specify starting index

  v.rfind(a)  - finds the last occurrence of N in Array or -1
  v.rfind(a,start) - specify starting index

  v[N]        - returns item at index N, if N is negative it is from last element

  v == v2     - returns true if counts and contents match

  foreach( T a; v ) ...         - standard foreach operator
  foreach( int n, T item; v )   - standard foreach operator with indices

  Array v.map( delegate bool ( T val ) check )    - calls dg(item)  for each item in Array
  Array v.filter( delegate bool( T val ) check)  - returns an Array of items which pass dg( item )
*/


private extern (C) void memmove( void* dest, void* src, uint num );

private import std.stdio;

class IndexException : Exception
{
  this( int index, int count, char[] where )
  {
    this.msg = format( "IndexException: attempted to access index %s, " ~
        "number of items %s.\nLocation:%s", index, count, where );
    super( msg );
  }

  this( int r1, int r2, int count, char[] where )
  {
    this.msg = format( "IndexException: attempted to access range %s-%s, " ~
        "number of items %s.\nLocation:%s", r1, r2, count, where );
    super( msg );
  }
}


private const int InitialSize = 8;


class Array(T)
{
  protected:
  T[]     data;
  uint    mCount=0;

  alias typeof(this)    TArray;

  invariant { assert( mCount<=data.length ); }


  // index conversion routine, all indices from API should be funneled thru this
  protected int   IndexOf( int n )
  {
    if( n<0 ) n = mCount + n;
    //debug writefln("Index at:", n, " max:", mCount );
    return n;
  }

  void ensure( uint amount )
  {
    int N = data.length;
    if ( N==0 ) N = InitialSize;
    while( N<amount ) N = N * 2;
    data.length = N;
  }


  public:
  this() { }

  this( T item )
  {
    ensure( 1 );
    push( item );
  }

  this( T[] items )
  {
     ensure(items.length);
     mCount = items.length;
     data[0..mCount] = items[0..mCount];
  }

  ~this() { data.length = 0; mCount=0; }

  uint  min()       { return 0; }
  uint  max()       { return mCount-1; }
  uint  capacity()  { return data.length; }
  uint  count()     { return this.mCount; }
  bool  empty()     { return this.mCount==0; }

  void  count( int n ) {
    data.length = n;
    mCount = n;
  }


  // just like the native array .dup property
  Array dup()
  {
    return new Array( this[].dup );
  }

  // clears elements of this Array, does not change capacity
  void clear()  { mCount = 0; }


  T         at( int N )
  {
    return data[IndexOf(N)];
  }


  // slice returns a copy of the array
  T[]       opSlice()
  {
    return data[0..mCount];
  }

  /*
    returns a portion of this Array, negative indexes allowed
    a[ -5 .. a.max ] would return the last 5 elements of the array
  */
  T[]       opSlice(int a, int b )
  {
    int from = IndexOf( a );
    int to = IndexOf( b );
    if ( from > to ) throw new IndexException( from, to, mCount, "array.d :: opSlice(a,b)");
    return data[from..to];
  }

  T[]      filter( bool delegate( T val ) dg )
  {
    TArray result = new TArray();
    foreach( T val ; data[0 .. mCount] )
    {
      if ( dg(val) ) result.append( val );
    }
    return result[];
  }


  void        map( bool delegate( inout T val ) dg )
  {
    for( int n=0; n<mCount; n++ ) if ( dg(data[n])==false ) break; ;
  }


  // passthru for native array reverse
  void reverse()
  {
    data[0..mCount].reverse;
  }


  // allocate an item, return a pointer to it
  // primarily for structs, which are pass by value
  T*    next()
  {
    ensure( mCount+1 );
    mCount ++;
    return data.ptr + (mCount-1);
  }


  // append item on the Array, return its index
  int   append(T item)
  {
    ensure( mCount+1 );
    data[mCount] = item;
    mCount++;
    return mCount-1;
  }

  // push N items on the Array
  void  append( T[] items )
  {
    uint sz = items.length;
    if ( sz==0 ) return;
    ensure( mCount + sz );
    data[mCount..mCount+sz] = items[0..sz];
    mCount += sz;
  }


  /*
    Some operations to allow streaming elements in and out
  */
  TArray  opShl( T[]  items ) { append( items ); return this; }
  TArray  opShl( Array items) { append( items.data ); return this; }
  TArray  opShl( T    item  ) { append( item ) ; return this; }
  TArray  opShr( out T item ) { item = pop(); return this; }


  bool  remove( T item )
  {
    int N = find( item );
    if ( N<0 ) return false;
    this.removeAt( N );
    return true;
  }

  void  removeAll( T item )
  {
    while( remove( item ) ) { ; }
  }


  // remove a single item from the Array
  T     removeAt( int N )
  {
    int index = IndexOf(N);
    T result = data[index];
    removeRange( index,index+1 );
    return result;
  }


  // remove items from..to ( this[to] is now in position this[from] )
  void removeRange( int from, int to )
  {
    int N1 = IndexOf(from);
    int N2 = IndexOf(to);
    if ( N1==N2 ) return; // no-op
    assert( N1<N2 );

    int NUM = mCount - N2  ; // N is number of items to move
    if ( NUM>0 ) {
      int nbytes = NUM * (T.sizeof);
      memmove( data.ptr + N1, data.ptr + N2, nbytes );
    }
    mCount -= (N2-N1);
  }



  // remove last item pushed
  T     pop()
  in { assert( mCount>0 ); }
  body
  {
    T ref = data[mCount-1];
    mCount--;
    return ref;
  }


  // remove N items from the Array
  void  drop( int n )
  in { assert(n>=0); }
  body
  {
    if ( n<mCount ) mCount = mCount-n;
    else clear();
  }

  // returns the first item pushed
  T   first()
  in { assert( mCount>0 ); }
  body
  {
    return data[0];
  }


  // returns the last item pushed
  T   last()
  in { assert( mCount>0 ); }
  body
  {
    return data[mCount-1];
  }

  // returns the oldest N elements
  T[]   first( int n )
  in { assert(n<=mCount); }
  body
  {
    return data[0..n];
  }


  // returns the recent N elements
  T[]   last( int n )
  in { assert(n<=mCount); }
  body
  {
    return data[mCount-n..mCount];
  }

  // aliases for the above two functions
  T[]  left( int n )    { return first(n); }
  T[]  right( int n )   { return last(n);  }


  // exchange the values at indices N1 and N2
  void swap( int N1, int N2 )
  {
    if ( N1<0 ) N1 = mCount + N1;
    if ( N2<0 ) N2 = mCount + N2;
    T tmp = data[N2];
    data[N2] = data[N1];
    data[N1] = tmp;
  }


  // aliases for using Array as a Stack
  T  peek()   { return data[ mCount-1 ]; }
  T  top()    { return data[ mCount-1 ]; }
  alias   append push;
  void    swap() { swap(-1,-2); }
  void    rot()  { swap(-1,-3); swap(-1,-2); }

  // other useful aliases
  alias   append add;
  T       tail() { return data[mCount-1]; }
  T       head() { return data[0]; }


  // inserts an item at position zero
  void insert( T val )
  {
    return this.insertAt( 0, val );
  }


  // inserts an item at the specified index, shifts remaining elements up by one
  void insertAt( int index, T val  )
  in { assert( index<=mCount ); }
  body
  {
    if ( index==mCount )
    {
      push( val );
      writefln("Insert at tail");
      return;
    }

    ensure(mCount+1);
    uint to_move = mCount - index;
    uint Nbytes = to_move * T.sizeof;

    memmove( data.ptr+index+1 , data.ptr+index , Nbytes );
    data[index] = val;
    mCount++;
  }


  // shifts all elements in the Array down by N items
  // the lower N items are removed
  void shift( int n )
  {
    removeRange( 0,n );
  }


  // Finds the first occurence of a value, or -1 if not found
  int     find( T val )
  {
    return find( val, 0 );
  }

  // finds the member of the Array which equals val, starting at specified index
  int     find( T val, int start )
  in { assert( start>=0 ); }
  body
  {
    if ( start>=mCount ) return -1;
    for( int n=start; n<mCount;n++ )
    {
      if ( data[n]==val ) return n;
    }
    return -1;
  }

  // looks for the specified value in reverse, beginning at the end of the array
  // returns -1 on fail
  int    rfind( T val )
  {
    return rfind( val, mCount-1 );
  }

  // perfoms a reverse search starting at the specified index, -1 on fail
  int    rfind( T val, int start )
  {
    if ( start > mCount ) start = mCount-1;
    for( int n=start; n>=0; n-- )
    {
      if ( data[n]==val ) return n;
    }
    return -1;
  }


  // standard array indexing
  // Note: negative indices are from the end of the Array
  // ( this[-1] = last, this[0]=first )
  T   opIndex( int n )
  {
    return at( n );
  }


  alias pointer ptr;

  // returns a pointer to an object, needed for in-place struct modification
  T*  pointer( int index = 0 )
  {
    int N = IndexOf(index);
    if ( (N<0) || (N>=mCount) )
      throw new IndexException( index,mCount, "Array.d :: pointer" );
    else return &data[N];
  }


  // standard array index assignment
  // Note: negative indices are from the end of the Array
  // ( this[-1] = last, this[0]=first )
  T opIndexAssign( T val, int index )
  in { assert( mCount>0 ); }
  body
  {
    alias std.string.toString str ;

    int N = index;
    if ( N<0 ) N = mCount + index;
    if( (N<0) || (N>=this.mCount))
    {
      throw new IndexException( index, mCount, "Array.d :: opIndexAssign" );
    }
    else
      data[N] = val;

    return val;
  }

  // Create a new Array concatenating with another Array
  T[] opCat( TArray other )
  {
    return data[0..mCount] ~ other;
  }

  // Create a new Array concatenating with native array
  T[]   opCat( T[] other )
  {
    return data[0..mCount] ~ other;
  }

  // Create a new Array concatenating with native array
  T[] opCat_r( T[] other )
  {
    return other ~ data[0..mCount];
  }


  // returns true if count and values are equal
  bool  opEquals( TArray other )
  {
    if ( other.count!=this.count ) return false;
    return data[0..mCount]==other[];
  }

  // returns true if count and values are equal
  bool  opEquals( T[] other )
  {
    if ( other.length != mCount ) return false;
    return data[0..mCount] == other[0..mCount];
  }

  // foreach over all items
  int opApply( int delegate(inout T) dg )
  {
    int result = 0;
		for (int i = 0; i < mCount; i++)
		{
		    result = dg(data[i]);
		    if (result)	break;
		}
		return result;
  }

  // foreach over items and with indices
  int opApply( int delegate(inout int index,inout T) dg )
  {
    int result = 0;
		for (int i = 0; i < mCount; i++)
		{
      int temp = i;                 // do not allow modification
		  result = dg(temp,data[i]);
		  if (result)	break;
		}
		return result;
  }
}


alias Array Stack;

alias Array!(char) String;


version( ArrayTest )
{
  import std.c.stdio;

  alias Array!(int)     IntArray;

  void main( char[][] arg )
  {
    Stack!(int) s = new Stack!(int);
    IntArray v ;

    try
    {
      printf("Begin \n");
      v = new IntArray();
      printf("Ctor() ok\n");

      static int[] arr = [ 1, 2 ];
      v = new IntArray( arr );
      writefln("Ctor(val[]) ok");

      v = new IntArray( 1 );
      writefln("Ctor(val) ok");

      test(v);
    }
    catch( Exception e )
    {
      writefln("Failed: %s", e.msg );
      if ( v is null ) writefln("Failed ctor of Array!");
      else foreach( int index,int val; v ) writefln( "v[%s]=%s", index, val );
    }

    String a = new String("Hello World");
    String b = new String( "Cruel" );
    String str = new String(a ~ b);
    assert( str == "Hello WorldCruel" );

    str = new String( a.first(6) ~ b );  // char[] ~ String
    str = new String( b ~ a.last(6) );  // String ~ char[]
    assert( str == "Cruel World" );

    bool testChar( char c ) { return c>='a' && c<='z'; }
    String s2 = new String( str.filter( &testChar ) );
    assert( s2 == "ruelorld" );
  }


  void dump( IntArray a )
  {
    foreach( int index,int val; a ) writefln( "v[%s]=%s", index, val );
  }

  void test(IntArray v)
  {
    v.clear();
    v << 1;

    assert( v.first == 1 );
    assert( v.first == v.last );
    assert( v.head == v[0] );
    assert( v.first == v[-1] );

    writefln("Testing push, pop..");
    v.push( 3 );
    assert( v.last == 3 );

    int val ;
    v >> val;
    assert( val == 3 );

    assert( v.count==1 );
    v << 3;
    assert( v[1] == 3 );
    v << 2;
    assert( v.last == 2 );
    v.insert( 150 );
    assert( v.first == 150 );
    v.insert( 200 ) ;
    assert( v[0]==200 );
    v.shift(2);

    static int[2] i2 = [ 4, 5 ];
    v << i2;
    assert( v.last == 5 );
    assert( v[ v.count-2 ] == 4 );
    v.drop( 3 );

    writefln("Testing insert..");
    v.insertAt( 1, 2 ); // put a 2 at position 1
    assert( v.first == 1 );
    assert( v[0]==1 );
    assert( v[1]==2 );
    assert( v[2]==3 );

    writefln("Testing first(N),last(N).." );
    assert( v.first(2)[1] == 2 );
    assert( v.last(2)[0] == 2 );
    assert( v.first(2)[0] == v.first );
    assert( v.first(3)[2] == v.last );

    writefln( "Testing reverse" );
    v.reverse();
    assert( v.first == 3 );
    assert( v.last == 1 );
    assert( v[1] == 2 );
    v.reverse(); // should be [1,2,3] now

    writefln( "Testing swap" );
    v.swap( 0,1 );
    assert( v.first == 2 );
    assert( v[1] == 1 );
    v.swap( 0,-1 );
    assert( v.last == 2 );
    assert( v.first == 3 );
    v.swap( -1, 1 );
    assert( v[1]==2 );
    assert( v[2]==1 );
    v.clear();
    v << 1 << 2 << 3;
    writefln( "start: %s,%s,%s", v[0], v[1], v[2] );
    assert( v.count==3 );
    v.rot();
    writefln( "rot: %s,%s,%s", v[0], v[1], v[2] );
    assert( v.last==2 );
    assert( v[-2]==1 );
    assert( v[0]==3 );
    v.swap();
    writefln( "swap: %s,%s,%s", v[0], v[1], v[2] );
    assert( v[1]==2 );
    assert( v[2]==1 );
    v.reverse();
    writefln( "reverse: %s,%s,%s", v[0], v[1], v[2] );

    writefln( "Testing equality, empty slice operator" );
    IntArray temp = new IntArray( v[] );
    assert( v.count == temp.count );
    assert( v == temp );

    int[] temp2 = temp[].dup.reverse;
    temp.reverse();
    assert( temp == temp2 );


    v.clear();
    assert( v.count==0 );


    writefln("Testing iteration");
    static int[5] temp3 = [ 100, 200, 300, 400, 500 ];
    v << temp3;

    uint sum = 0;
    foreach( int n ; v ) sum += n ;
    assert( sum == 1500 );


    writefln("Testing remove" );
    v.removeRange( 2,4 );
    assert( v.count==3 );
    assert( v[0]==100 );
    assert( v[1]==200 );
    assert( v[2]==500 );
    v.removeAt(1);
    assert( v.count==2 );
    assert( v[0]==100 );
    assert( v[1]==500 );
    v.shift(1);
    assert( v[0]==500 );
    assert( v.count==1 );

    v << 200 << 400 ;
    assert( v.count==3 );
    v.swap(0,1);
    v.swap(1,2);

    writefln( "Testing dup" );
    IntArray v2 = v.dup;
    assert( v2==v );
    v2[0] = 50 ;
    assert( v[0] != v2[0] );
    v2.clear();
    assert( v != v2 );
    v.clear();

    writefln("Testing removeRange" );
    v << temp3;

    //writefln( "Sum = ", Sum!(double).of( v ) );

    v.removeRange( 1, -2 ); // should be = [ 100, 400, 500 ]
    assert( v.count == 3 );
    v.removeAt(-1);
    assert( v.count == 2 );
    assert( v.last == 400 );
    writefln("Passed tests" );
  }
}



