module vecmath;

// Vector math library, implemented as a stack

import array;

private import std.math;
private import std.stdio;


// floating point comparison with specified accuracy
private template EQV( uint decimals )
{
  const double epsilon = 1.0 / ( 10 * decimals );

  bool equal( double a, double b )
  {
    return fabs( b-a ) < epsilon;
  }
}


private double Round( double val )
{
  long tmp = cast(long)( val * 10000 );
  return (cast(double)( tmp ) / 10000) ;
}



vec4  vec( double a, double b, double c, double d = 0 )
{
  vec4 result;
  result.set( a,b,c,d );
  return result;
}


double radians( double deg )
{
  return PI * ( deg / 180.0 );
}

double degrees( double rad )
{
  return ( rad / PI ) * 180.0;
}



// Basic vector structure
struct vec4
{
  float v[4] =  0;
  float x() { return v[0]; }
  float x(float f ) { return v[0]=f; }
  float y() { return v[1]; }
  float y(float f ) { return v[1]=f; }
  float z() { return v[2]; }
  float z(float f ) { return v[2]=f; }
  float w() { return v[3]; }
  float w(float f ) { return v[3]=f; }

  void opCall( double d ) { set( d ); }
  void opCall( double xx, double yy, double zz, double ww = 0 ) { set( xx,yy,zz,ww ); }

  void set( double d ) { v[] = cast(float)d; }
  void set( double x, double y, double z, double w = 0 )
  {
    v[0] = x;
    v[1] = y;
    v[2] = z;
    v[3] = w;
  }

  float[] opSlice()   { return v[]; }

  float opIndex( int n ) { return v[n]; }
  vec4  opNeg() { vec4 r; r.set( -x, -y, -z, -w ); return r; }
  vec4  opMul(double s) { vec4 r; r.set( s*x, s*y, s*z, s*w ); return r; }
  vec4  opAdd(vec4 v) { vec4 r; r.set( x+v.x, y+v.y, z+v.z, w+v.w ); return r; }
  vec4  opSub(vec4 v) { vec4 r; r.set( x-v.x, y-v.y, z-v.z, w-v.w ); return r; }
  double opIndexAssign( double d, int n ) { return v[n] = d; }
}



private double M( int row, int col, double[] mat ) { return mat[ row + ( col * 4 ) ]; }


// Quaternion structure
alias vec4  quat;



// -----------------------------------------------------
// Encapsulates vector operations in a forth-like Object
// -----------------------------------------------------
class VecStack : Array!(vec4)
{
  // ----------------------------------------------------------------
  // VECTOR functions
  // ----------------------------------------------------------------
  void push( float[] v )
  {
    vec4 t;
    int max = (v.length >= 4) ? 4 : v.length;
    t.v[0..max] = v[0..max];
    super.push( t );
  }

  void push( vec4[] arr )
  {
    super.push( arr );
  }

  void push( vec4 v ) { super.push( v ); }
  void push() { vec4 v; super.push( v ); }


  void push( double x, double y, double z=0, double w =0 )
  {
    vec4* v = super.next();
    v.v[0] = x;
    v.v[1] = y;
    v.v[2] = z;
    v.v[3] = w;
  }

  private alias EQV!(3).equal equalf;

  bool  equal() // ( a b -- )
  {
    vec4 b = pop();
    vec4 a = pop();
    return  equalf( a.x, b.x ) && equalf( a.y, b.y ) &&
            equalf( a.z, b.z ) && equalf( a.w, b.w );
  }


  bool  equal( double a, double b, double c ) // ( a -- a )
  {
    return equalf( top.v[0], a ) &&
           equalf( top.v[1], b ) &&
           equalf( top.v[2], c );
  }


  bool  equal( double a, double b, double c, double d ) // ( a -- a )
  {
    return equalf( top.v[0], a ) &&
           equalf( top.v[1], b ) &&
           equalf( top.v[2], c ) &&
           equalf( top.v[3], d );
  }

  double dot() // ( a b -- )
  {
    assert( this.count >1 );
    vec4 b = pop();
    vec4 a = pop();
    return a[0] * b[0] + a[1] * b[1] + a[2] * b[2] + a[3] * b[3];
  }


  double magnitude()
  {
    super.push( this.last );
    return sqrt( dot() );
  }

  // returns inner angle in radians
  double inner_angle() // ( v1, v2 -- )
  {
    super.push( last(2) );
    double Lb = magnitude();
    double La = magnitude();
    return acos ( dot() ) / ( La * Lb );
  }


  void  negate() // ( v -- -v )
  {
    scale( -1.0 );
  }

  void add() // ( a b -- v )
  {
    vec4 b = pop();
    vec4 a = pop();

    vec4* c= next();
    c.v[0] = a.x + b.x;
    c.v[1] = a.y + b.y;
    c.v[2] = a.z + b.z;
    c.v[3] = a.w + b.w;
  }

  void sub() // ( a b -- v )
  {
    negate();
    add();
  }


  void  add( double d )
  {
    vec4* a = pointer( -1 );
    a.v[0] = a.v[0] + d;
    a.v[1] = a.v[1] + d;
    a.v[2] = a.v[2] + d;
    a.v[3] = a.v[3] + d;
  }

  void  sub( double d )
  {
    this.add( -d );
  }

  void  scale( double d )
  {
    vec4* vec = pointer( this.max );
    vec.v[0] *= d;
    vec.v[1] *= d;
    vec.v[2] *= d;
    vec.v[3] *= d;
  }

  void  cross()
  {
    vec4 b = pop();
    vec4 a = pop();
    vec4* c = next();
    c.v[0] = a.y*b.z - a.z*b.y;
    c.v[1] = a.z*b.x - a.x*b.z;
    c.v[2] = a.x*b.y - a.y*b.x;
    c.v[3] = 0;
  }

  void normalize()
  {
    this.normalize( 1.0 );
  }

  void normalize( double len )
  {
		super.append( this.last );
		double mag = this.magnitude();
		if ( mag==0 ) return;
		this.scale( len / mag );
  }


  private vec4 transform( double[16] matrix, vec4 v )
  {
    vec4 r;
    r.x = v.x * matrix[0] +
          v.y * matrix[4] +
          v.z * matrix[8] +  matrix[12];

    r.y = v.x * matrix[1] +
          v.y * matrix[5] +
          v.z * matrix[9] +  matrix[13];

    r.z = v.x * matrix[2] +
          v.y * matrix[6] +
          v.z * matrix[10] +  matrix[14];
    r.w = v.w;
    return r;
  }


  void transform( double[16] matrix )
  {
    push( transform( matrix, pop() ) );
  }


  void transformAll( double[16] matrix )
  {
    for( int N=0; N<this.count; N++ )
    {
      this[N] = transform( matrix, this[N] );
    }
  }



  // ----------------------------------------------------------------
  // Quaternion functions
  // ----------------------------------------------------------------
  void quat_mul()  // ( q1 q2 -- qr )
  {
    vec4 q2 = pop();
    vec4 q1 = pop();
    vec4 r;
    r.w = q1.w * q2.w - q1.x * q2.x - q1.y * q2.y - q1.z * q2.z ;
    r.x = q1.w * q2.x + q1.x * q2.w + q1.y * q2.z - q1.z * q2.y ;
    r.y = q1.w * q2.y - q1.x * q2.z + q1.y * q2.w + q1.z * q2.x ;
    r.z = q1.w * q2.z + q1.x * q2.y - q1.y * q2.x + q1.z * q2.w ;
    push( r.x, r.y, r.z, r.w );
  }


  void quat_identity()
  {
    vec4 * v = super.next();
    v.x = v.y = v.z = 0;
    v.w = 1.0;
  }

  void quat_conjugate() // ( q -- 'q )
  {
    quat q = pop();
    push( -q.x, -q.y, -q.z, -q.w );
  }


  double quat_normal() // ( q -- )
  {
    return dot();
  }


  void quat_inverse() //( q -- 'q )
  {
    super.push( this.last );
    double N = quat_normal();
    quat_conjugate();
    this.scale( 1.0 / N );
  }



  void axis_angle_to_quat( double radians ) // ( axis -- quat )
  {
    normalize();
    vec4 axis = pop();
    radians = radians * 0.5;
    float C = cos( radians );
    float S = sin( radians );
    push( axis.x * S, axis.y * S, axis.z * S, C );
    normalize();
  }


  /*
    Finds a quaternion which will rotate from starting axis A to Ending Axis B
    eg. if A=(1,0,0) and B=(0,1,0) then result will be
        Quat( Axis angle of (0,0,1) and angle = 90 degrees )
  */
  void axis_axis_to_quat()  // ( a b -- Q )
  {
    push( this.last(2) );   // ( a b a b )
    cross();                // ( a b cross -- )
    vec4  pivot = pop(); // ( a b )
    swap( -1, -2 );      // ( b a )
    normalize();         // ( b na )
    swap( -1, -2 );      // ( na b )
    normalize();         // ( na nb )
    double angle = acos( dot() ); // ()
    push( pivot );       // ( pivot -- )
    axis_angle_to_quat( angle );  // ( q )
  }


  double quat_to_axis_angle() // ( q -- axis ) returns angle in radians
  {
    vec4  q = pop();
    push( q.x, q.y, q.z, 0 );
    double scale = magnitude();
    double angle = 2.0 * acos( q.w );
    double x = q.x / scale;
    double y = q.y / scale;
    double z = q.z / scale;
    push( x, y, z, 0 );
    return angle;
  }




  void quat_to_matrix( float[16] m )
  {
    double[16] temp;
    quat_to_matrix( temp );
    foreach( int n, double d; temp ) m[n] = cast(float)d;
  }


  /*
   [ 1 - 2y2 - 2z2    2xy - 2wz      2xz + 2wy
     2xy + 2wz    1 - 2x2 - 2z2    2yz - 2wx
     2xz - 2wy      2yz + 2wx    1 - 2x2 - 2y2 ]
  */
  final void quat_to_matrix( double[16] M ) // ( q -- )
  {
    normalize();
    quat v = pop();
    double xx = v.x*v.x;
    double xz = v.x*v.z;
    double xy = v.x*v.y;
    double xw = v.x*v.w;
    double yy = v.y*v.y;
    double yz = v.y*v.z;
    double yw = v.y*v.w;
    double zz = v.z*v.z;
    double zw = v.z*v.w;

    M[] = 0;
    M[15] = 1;

    M[0] = 1 - 2 * (yy + zz) ;
    M[1] =  2 * (xy + zw) ;
    M[2] =  2 * (xz - yw) ;

    M[4] = 2 * (xy - zw) ;
    M[5] =  1 - 2 * (xx + zz) ;
    M[6] =  2 * (yz + xw) ;

    M[8] = 2 * (xz + yw) ;
    M[9] =  2 * (yz - xw) ;
    M[10] = 1 - 2 * (xx + yy) ;
  }


  // X Axis of a quaternion, row 0 in OpenGL
  void quat_push_xaxis() // q -- q v
  {
    quat q = peek();
    vec4 v ;
    v.x = 1 - 2 * (q.y*q.y + q.z*q.z) ;
    v.y = 2 * (q.x*q.y + q.z*q.w) ;
    v.z = 2 * (q.x*q.z - q.y*q.w) ;
    v.w = 0;
    push( v );
  }

  // Y Axis of a quaternion, row 1 in OpenGL
  void quat_push_yaxis() // q -- q v
  {
    quat q = peek();
    vec4 v ;
    v.x = 2 * (q.x*q.y - q.z*q.w) ;
    v.y = 1 - 2 * (q.x*q.x + q.z*q.z) ;
    v.z = 2 * (q.y*q.z + q.x*q.w) ;
    v.w = 0;
    push( v );
  }

  // Z Axis of a quaternion, row 2 in OpenGL
  void quat_push_zaxis() // q -- q v
  {
     quat q = peek();
     vec4 v ;
     v.x = 2 * ( q.x * q.z + q.y * q.w );
     v.y = 2 * ( q.y * q.z - q.x * q.w );
     v.z = 1 - 2 * ( q.x * q.x + q.y * q.y );
     v.w = 0;
     push( v );
  }



  // unfinished
  void matrix_to_quat( double[16] m )
  {
    double T = m[0] + m[5] + m[10] + 1;

    int largest( double a, double b, double c )
    {
      if ( a > b ) { if ( a > c ) return 1; else return 3; }
      // a is out of the running
      if ( b > c ) return 2; else return 3;
    }

    quat* q = this.next();

    if ( T > 0.0 )
    {
      double S = 0.5 / sqrt(T);
      q.w = 0.25 / S;
      q.x = ( m[9] - m[6] ) * S;
      q.y = ( m[2] - m[8] ) * S;
      q.z = ( m[4] - m[1] ) * S;
      return;
    }
    else
    {
      double S;
      switch( largest( m[0], m[5], m[10] ) )
      {
        case 1:
          S  = sqrt( 1.0 + m[0] - m[5] - m[10] ) * 2;
          q.x = 0.5 / S;
          q.y = (m[1] + m[4] ) / S;
          q.z = (m[2] + m[8] ) / S;
          q.w = (m[6] + m[9] ) / S;
          return ;

        case 2:
          S  = sqrt( 1.0 + m[5] - m[0] - m[10] ) * 2;
          q.x = (m[1] + m[4] ) / S;
          q.y = 0.5 / S;
          q.z = (m[6] + m[9] ) / S;
          q.w = (m[2] + m[8] ) / S;
          return;

        case 3:
        default:
          S  = sqrt( 1.0 + m[10] - m[0] - m[5] ) * 2;

          q.x = (m[2] + m[8] ) / S;
          q.y = (m[6] + m[9] ) / S;
          q.z = 0.5 / S;
          q.w = (m[1] + m[4] ) / S;
          return;
      }
    }

    throw new Exception( "vecmath.d :: matrix_to_quat not implemented!!" );
  }

  // ----------------------------------------------------------------
  // Euler functions
  // ----------------------------------------------------------------


  void quat_to_euler_angles() // ( q -- <rx ry rz> );
  {
    vec4 q = pop();
    double X=0, Y=0, Z = 0;

  	double test = q.x*q.y + q.z*q.w;
  	if (test > 0.499)
  	{
  	  Z = std.math.PI/2;
  		Y = 2 * atan2(q.x,q.w);
  		X = 0;
  	}
  	else if (test < -0.499)
  	{
  	  Z = - std.math.PI/2;
  		Y = -2 * atan2(q.x,q.w);
  		X = 0;
  	}
  	else
  	{
      double sqx = q.x*q.x;
      double sqy = q.y*q.y;
      double sqz = q.z*q.z;
      Z = asin(2*test);
      Y = atan2(2*q.y*q.w-2*q.x*q.z , 1 - 2*sqy - 2*sqz);
  	  X = atan2(2*q.x*q.w-2*q.y*q.z , 1 - 2*sqx - 2*sqz);
	  }

  	this.push( X, Y, Z, 0 );
  }

  /*
  Qx = [ cos(a/2), (sin(a/2), 0, 0)]
  Qy = [ cos(b/2), (0, sin(b/2), 0)]
  Qz = [ cos(c/2), (0, 0, sin(c/2))]
  */
  void euler_to_quat()
  {
    vec4 e = pop() * 0.5;

    quat q1, q2, q3;
    q1.set( sin(e.x), 0, 0, cos(e.x) );
    q2.set( 0, sin(e.y), 0, cos(e.y) );
    q3.set( 0, 0, sin(e.z), cos(e.z) );

    push( q1 );
    push( q2 );
    push( q3 );
    quat_mul();
    quat_mul();
  }


  /*
  void euler_to_matrix( double[16] M )
  {
    vec4 e = pop();
    push( 1, 0, 0 );
    axis_angle_to_quat( e.x );
    push( 0, 1, 0 );
    axis_angle_to_quat( e.y );
    quat_mul();
    push( 0, 0, 1 );
    axis_angle_to_quat( e.z );
    quat_mul();
    quat_to_matrix( M );
  }
  */


  void euler_to_matrix( double[16] M )
  {
    this.euler_to_quat();
    this.quat_to_matrix( M );
  }



  // convert a matrix to Euler angles
  void matrix_to_euler( double[16] mat ) // -- <rx, ry, rz>
  {
    double rx, ry, rz;
  	if ( M(1,0,mat) > 0.998) { // singularity at north pole
  		ry = atan2( M(0,2,mat) ,M(2,2,mat) );
  		rx = PI/2;
  		rz = 0;
  	}
  	if (M(1,0,mat) < -0.998) { // singularity at south pole
  		ry = atan2(M(0,2,mat) ,M(2,2,mat) );
  		rx = -PI/2;
  		rz = 0;
  	}
  	else
  	{
    	ry = atan2(-M(2,0,mat) ,M(0,0,mat) );
    	rz = atan2(-M(1,2,mat) ,M(1,1,mat) );
    	rx = asin(M(1,0,mat));
  	}
  	push( rx, ry, rz, 0 );
  }



  // ----------------------------------------------------------------
  // Debugging / printing functions
  // ----------------------------------------------------------------


  void print_vector( char[] msg, int index=-1 )
  {
    writef( "%s:", msg );
    print_vector( index );
  }

  void print_vector(int index = -1) //( v -- v )
  {
    vec4 vec = at(index) ;
    with( vec ) writefln("<%f,%f,%f,%f>", Round(v[0]), Round(v[1]), Round(v[2]), Round(v[3]) );
  }

  void print_angles( char[] msg, int index=-1 )
  {
    writef( "%s:", msg );
    vec4 v = at(index);
    with(v) writefln("<%f,%f,%f,%f>", degrees(v[0]), degrees(v[1]), degrees(v[2]), degrees(v[3]) );
  }

}

// matrix math routines
template Matrix(T)
{
  private int at(int r,int c) { return  (c*4) + r;}

  void Identity( T[16] m )
  {
    m[] = 0;
    m[0] = m[5] = m[10] = m[15] = 1;
  }


  void  Inverse( T[16] a, T[16] dest )
  {
    T[16] r;
    for( int i=0; i<4; i++ )
    {
      for( int j=0; j<4; j++ )
      {
        if ( i==j ) continue;
        r[ at(i,j) ] = a[ at(j,i) ];
      }
    }
    dest[0..16] = r[0..16];
  }

  void  Multiply( T[16] a, T[16] b, T[16] dest )
  {
    double Sum( int r, int c )
    {
      int n = c * 4;
      return a[r] * b[ n ] +
             a[r+4] * b[n+1] +
             a[r+8] * b[n+2] +
             a[r+12] * b[n+3];
    }

    for( int row=0; row<4; row++ )
    for( int col=0; col<4; col++ )
    {
      dest[ at(row,col) ] = Sum( row, col );
    }
  }

  void Transform( T[16] M, T[] V )
  {
    double MulSum( int row )
    {
      return M[row] * V[0] + M[row+4] * V[1] +
              M[row+8] * V[2] + M[row+12];
    }

    V[0] = MulSum( 0 );
    V[1] = MulSum( 1 );
    V[2] = MulSum( 2 );
  }

  void Print( T[16] matrix )
  {
    writefln("\nMatrix:" );
    for( int i=0; i<4; i++ )
      writefln("<%s,%s,%s,%s>", Round(matrix[i]), Round(matrix[i+4]),
                                Round(matrix[i+8]), Round(matrix[i+12]) );
  }
}


void main( char[][] arg )
{
  VecStack vm = new VecStack();
  vm.push( 0, 0, radians(90) );
  double[16] M;
  vm.euler_to_matrix( M );

  // Here vm is now NULL !!!!!

  writefln( "Euler(0,0,90 degrees) -> Matrix" );
  Matrix!(double).Print( M );
  assert( vm.count==0 ); // access violation
}
